/* $Id: types.h,v 1.15 2009/01/24 15:33:55 mit-sato Exp $ */

#ifndef __TYPES__
#define __TYPES__

#include <stdio.h>
#include <stdlib.h>
#include <t_gc.h>
#include <setjmp.h>
#include "cell.h"
#include "hash.h"
#include "array.h"
#include "bulk.h"

#define    NIL			0	/* nil */
#define    SYMBOL  		1	/* symbol */
#define	   REF			2	/* reference (e.g. $XX) */
#define    LIST	  		3	/* list (e.g. (...)) */
#define    INTEGER  		4	/* 64bit signed integer */
#define    REAL	  		5	/* 64bit real */
#define    STRING		6	/* string (e.g. "...") */
#define    SCRIPT		7	/* script is some statements */
#define    STATEMENT		8	/* statement (e.g. ...;) */
#define    BLOCK		9	/* block (e.g. {...} */
#define    EVAL	  		10	/* eval (e.g. [...] */
#define    NATIVE		11	/* C native code, no expression */
#define    OBJECT		12	/* a object */
#define    EXCEPTION		13	/* throwable object */
#define    CLOSURE		14	/* pending BLOCK */
#define    FUNC	  		15	/* function object */
#define	   CONTROL		16	/* control object (e.g. return, break, or continue) */
#define    CONTAINER		17	/* C object container */
#define    GETMACRO		18	/* GET Macro */
#define    ALIAS		19	/* variable alias */
#define    RQUOTE		20	/* reguler expresion string */
#define	   INITMACRO		21	/* INIT Macro */
#define    __TYPE_LAST__	22


#define S_NIL	"nil"

#define IS_LIST_NULL(l)	((l==NULL)?1:((NULL==l->u.list.item)&&(NULL==l->u.list.nextp)))

#define TAG_MASK		(0x000000ff)
#define GET_TAG(p)		((p==NULL)?-1:(TAG_MASK & (p->tag)))
#define TAG_NAMED_MASK		(0x00008000)
#define TAG_SWITCH_MASK		(0x00004000)
#define IS_NAMED_SYM(p)		((p==NULL)?0:(TAG_NAMED_MASK & (p->tag)))
#define IS_SWITCH_SYM(p)	((p==NULL)?0:(TAG_SWITCH_MASK & (p->tag)))
#define SET_SCRIPT_ID(p,i)	(p->tag|=(i<<16))
#define GET_SCRIPT_ID(p)	(int)((p->tag&0xffff0000)>>16)

#define IS_TOY_OBJECT(p)	((GET_TAG(p)>=NIL)&&(GET_TAG(p)<__TYPE_LAST__))
#define SELF_HASH(i)		(i->obj_stack[i->cur_obj_stack]->cur_object_slot)
#define SELF(i)			(i->obj_stack[i->cur_obj_stack]->self)
#define SELF_OBJ(i)		(i->obj_stack[i->cur_obj_stack]->cur_object)

typedef struct _toy_func_env {
    Hash *localvar;
    Hash *closurevar;
    struct _toy_type *statement;
    int line;
    int script_id;
    struct _toy_type *object_name;
    struct _toy_type *func_name;
} Toy_Func_Env;

typedef struct _toy_obj_env {
    struct _hash *cur_object_slot;
    struct _toy_type *cur_object;
    struct _toy_type *self;
} Toy_Obj_Env;

typedef struct _toy_env {
    struct _hash *localvar;
    Toy_Obj_Env *object_env;
} Toy_Env;

struct _toy_argspec {
    struct _toy_type *list;
    int posarg_len;
    Array *posarg_array;
    Hash *namedarg;
} Toy_Argspec;

typedef struct _toy_interp {
    /* interpriter name */
    char *name;

    /* stack size */
    int stack_size;

    /* some stacks */
    int cur_func_stack;
    Toy_Func_Env **func_stack;

    int cur_obj_stack;
    Toy_Obj_Env **obj_stack;

    /* world */
    struct _hash *funcs;
    struct _hash *classes;
    struct _hash *globals;
    struct _hash *scripts;

    /* execute script line */
    int line;

    /* execute statement */
    struct _toy_type *statement;

    /* execute script id */
    int script_id;

    /* object name */
    struct _toy_type *object_name;

    /* func name */
    struct _toy_type *func_name;
    

} Toy_Interp;

/*
 * TAG format:
 *       FEDCBA9876543210
 *      +----------------+
 *   +2 |   script id    |
 *      +----------------+
 *   +0 |@#      |  TYPE |
 *      +----------------+
 *
 *	@: named arg flag
 *	#: switch arg flag
 */
typedef struct _toy_type {
    int tag;

    union _u {
	/* NIL */

	/* SYMBOL */
	struct _symbol {
	    Cell *cell;
	    unsigned int hash_index;
	} symbol;

	/* REF */
	struct _ref {
	    Cell *cell;
	    unsigned int hash_index;
	} ref;

	/* LIST */
	struct _list {
	    struct _toy_type *nextp;
	    struct _toy_type *item;
	} list;

	/* INTEGER */
	long long int integer;

	/* REAL */
	double real;

	/* STRING */
	Cell *string;

	/* SCRIPT */
	struct _toy_type *statement_list;
	
	/* STATEMENT */
	struct _statement {
	    struct _toy_type *item_list;
	    int line;
	} statement;
	
	/* BLOCK */
	struct _toy_type *block_body;		/* point to _toy_type->script */

	/* EVAL */
	struct _toy_type *eval_body;		/* point to _toy_type->script */

	/* NATIVE */
	struct _native {
	    struct _toy_type* (*cfunc)(
		Toy_Interp* interp,
		struct _toy_type* posargs,	/* pint to _toy_type->list */
		struct _hash* namedargs,
		int arglen
		);
	    struct _toy_type *argspec_list;
	} native;

	/* OBJECT */
	struct _object {
	    struct _hash *slots;
	    struct _toy_type *delegate_list;
#ifdef HAS_MCACHE
	    struct _hash *mcache;
#endif
	} object;

	/* EXCEPTION */
	struct _exception {
	    Cell *code;
	    struct _toy_type *msg_list;
	} exception;

	/* CLOSURE */
	struct _closure {
	    struct _toy_type *block_body;	/* point to _toy_type->script */
	    Toy_Env *env;
	} closure;

	/* FUNC */
	struct _func {
	    struct _toy_argspec *argspec;
	    struct _toy_type *closure;
	} func;

	struct _control {
	    int code;
	    struct _toy_type *ret_value;
	} control;

	/* CONTAINER */
	void *container;

	/* HASH */
	struct _hash *hash;

	/* ARRAY */
	struct _array *array;

	/* GET Macro */
	struct _getmacro {
	    struct _toy_type *obj;
	    struct _toy_type *para;
	} getmacro;

	/* INIT Macro */
	struct _initmacro {
	    struct _toy_type *class;
	    struct _toy_type *param;
	} initmacro;

	/* ALIAS */
	struct _alias {
	    struct _hash *slot;
	    struct _toy_type *key;
	} alias;

	/* RQUOTE */
	Cell *rquote;

    } u;
} Toy_Type;

typedef struct _toy_script {
    int id;
    Toy_Type *path;
    Bulk *src;
    Toy_Type *script;
} Toy_Script;


Toy_Type*	new_nil();
Toy_Type*	new_symbol(char *atom);
Toy_Type*	new_ref(char *ref);
Toy_Type*	new_list(Toy_Type *item);
Toy_Type*	new_cons(Toy_Type *car, Toy_Type *cdr);
Toy_Type*	list_append(Toy_Type *list, Toy_Type *item);
Toy_Type*	list_next(Toy_Type *list);
Toy_Type*	list_get_item(Toy_Type *list);
int		list_length(Toy_Type *list);
Toy_Type*	new_integer(long long int integer);
Toy_Type*	new_real(double real);
Toy_Type*	new_string_str(char *string);
Toy_Type*	new_string_cell(Cell *string);
Toy_Type*	new_script(Toy_Type *statement_list);
Toy_Type*	new_statement(Toy_Type *item_list, int line);
Toy_Type*	new_block(Toy_Type *block_body);
Toy_Type*	new_eval(Toy_Type *eval_body);
Toy_Type*	new_native(Toy_Type* (*cfunc)(Toy_Interp*, Toy_Type*, struct _hash*, int arglen),
			   Toy_Type *argspec_list);
Toy_Type*	new_object(char* name, struct _hash* slots, Toy_Type *delegate_list);
Toy_Type*	new_exception(char *code, char *desc, Toy_Interp* interp);
Toy_Type*	new_closure(Toy_Type *block_body, Toy_Env* env, int script_id);
Toy_Type*	new_func(Toy_Type *argspec_list, int posarglen, Array *posargarray, Hash *namedarg,
			 Toy_Type *closure);
Toy_Type*	new_control(int code, Toy_Type *ret_value);
Toy_Type*	new_container(void *container);
Toy_Type*	new_getmacro(Toy_Type *obj, Toy_Type *para);
Toy_Type*	new_initmacro(Toy_Type *obj, Toy_Type *param);
Toy_Type*	new_alias(struct _hash *slot, Toy_Type *key);
Toy_Type*	new_rquote(char *string);

Toy_Type*	toy_clone(Toy_Type *obj);
char*		toy_get_type_str(Toy_Type *obj);
char*		to_string(Toy_Type *obj);
char*		to_print(Toy_Type *obj);

#define	list_next(l)		((l==NULL)?NULL:((GET_TAG(l)==LIST)?l->u.list.nextp:NULL))
#define list_get_item(l)	((l==NULL)?NULL:((GET_TAG(l)==LIST)?l->u.list.item:NULL))
#define list_set_car(l,v)	((l==NULL)?NULL:((GET_TAG(l)==LIST)?l->u.list.item=v:NULL))
#define list_set_cdr(l,v)	((l==NULL)?NULL:((GET_TAG(l)==LIST)?l->u.list.nextp=v:NULL))

#endif /* __TYPES__ */
